from dataclasses import dataclass, field
from typing import Mapping, List, Any
from datetime import datetime
import logging
import pandas as pd
import glob
import numpy as np
import logging
import os
from collections import OrderedDict

import nrrd
import vtk
import vedo
from vtk.util.numpy_support import numpy_to_vtk

from iblviewer.collection import Collection
import iblviewer.objects as obj
import iblviewer.utils as utils


@dataclass
class VolumeModel:
    RAW = 'raw'
    SEGMENTED = 'segmented'
    
    NORMALIZED_SUFFIX = '_norm'
    DATA_TYPE = {RAW:0, SEGMENTED:1}
    PREFIX = 'Volume'
    __count = 0

    def unique_name():
        VolumeModel.__count += 1 
        return f'{VolumeModel.PREFIX}_{VolumeModel.__count}'

    name: str = field(default_factory=unique_name)
    file_path: str = None

    scalars: Collection = field(default_factory=Collection)
    axes: List = field(default_factory=lambda: [1, 1, 1])
    data_min: float = None
    data_max: float = None
    data_map_step: float = 1.0
    data: np.ndarray = None
    data_type: str = RAW
    resolution: int = 1
    # Default units are microns.
    units: float = 1e-06
    base_color_map: Any = None

    # At IBL, volume mappings are used from ibllib: ibllib.atlas.regions.mappings
    mapping_name: str = None
    lateralized: bool = False
    # Mapping function. If None, the volume will be given as it is.
    mapping: Any = None

    luts: Collection = field(default_factory=Collection)
    slicers: Collection = field(default_factory=Collection)
    isosurfaces: Collection = field(default_factory=Collection)

    interactive_subsampling: bool = True
    volume_visible: bool = True
    slices_visible: bool = True

    transpose_shape: Any = None
    dimensions: np.ndarray = np.zeros(3).astype(float)
    center: np.ndarray = np.zeros(3).astype(float)

    def compute_size(self):
        """
        Compute volume size
        """
        if self.data is None:
            return
        self.dimensions = np.array(self.data.shape)[:3]
        if self.resolution is None:
            return
        self.resolution = int(self.resolution) # TODO: move this to constructor or init
        self.dimensions *= self.resolution
        self.center = np.ones(3) * self.resolution / 2 + self.dimensions / 2

    def compute_range(self, force=False):
        """
        Compute min and max range in the volume
        :return: Min and max values
        """
        if self.data_min is not None and self.data_max is not None and not force:
            return self.data_min, self.data_max
        self.data_min = np.min(self.data)
        self.data_max = np.max(self.data)
        #print('Volume min-max', self.data_min, self.data_max)
        return self.data_min, self.data_max

    def guess_volume_type(self):
        """
        Infer the volume type when it was not specified by the user.
        We assume here that typical values between -1 and 1 are raw volumes.
        """
        if self.data_type is None:
            if self.data_min is None or self.data_max is None:
                self.compute_range()
            if self.data_min >= -1 and self.data_max <= 1:
                guess = VolumeModel.RAW
            else:
                guess = VolumeModel.SEGMENTED
            self.data_type = guess

    def is_segmented(self, auto_guess=True):
        """
        Get whether current volume/image is segmented
        :return: Boolean
        """
        if self.data_type is None and auto_guess:
            self.guess_volume_type()
        return self.data_type == VolumeModel.SEGMENTED

    def read_volume(self, file_path):
        """
        Read local volume. Downloads the file first if it's remote.
        :param file_path: Volume path
        :return: 3D array
        """
        if file_path.startswith('http') or file_path.startswith('ftp'):
            downloaded_temp_file_path = vedo.download(file_path, verbose=False)
            if file_path.endswith('nrrd'):
                data, header = nrrd.read(downloaded_temp_file_path)
            else:
                data = vedo.loadImageData(downloaded_temp_file_path)
        else:
            if file_path.endswith('nrrd'):
                data, header = nrrd.read(file_path, index_order='C')
            else:
                data = vedo.loadImageData(file_path)
        return data

    def load_volume(self, file_path, remap_scalars=False, mapping=None, make_current=True):
        """
        Load a volume data file. Supports NRRD and many other formats thanks to vedo/VTK
        :param file_path: Volume file path. Could support other file types easily.
        :param remap_scalars: Whether scalar values in the volume are replaced by 
            their row id from a mapping that stores. This is necessary in the case of segmented
            volumes with regions that have a discontinuous id.
        :param mapping: Pandas Series or a Dictionary
        :param make_current: Set the volume data as the current one
        :return: 3D array
        """
        data = None
        if not remap_scalars or mapping is None:
            data = self.import_volume(file_path)
        else:
            time = datetime.now()
            new_file_path = utils.change_file_name(file_path, None, None, VolumeModel.NORMALIZED_SUFFIX)
            if os.path.exists(new_file_path):
                data = self.import_volume(new_file_path)
            else:
                data = self.import_volume(file_path)
                data, mapping = self.remap_slow(data, mapping, new_file_path)
                logging.info('Remapped scalar values in: ' + str(utils.time_diff(time)) + 's')
            '''
            if volume is not None:
                logging.info('Opened atlas ' + new_file_path + ' in ' + str(utils.time_diff(time)) + 's')
                min_value, max_value = np.amin(data), np.amax(data)
                logging.info('Min max scalar values in volume ' + str(min_value) + ' -> ' + str(max_value))
            else:
                logging.error('Failed to open atlas ' + new_file_path)
            '''
        if make_current and data is not None:
            self.data = data
        return data, mapping

    def transpose(self, shape=None):
        """
        Transpose the volume for visualization in VTK
        :param shape: The new shape. If None, will default to self.transpose_shape
        """
        if shape is None:
            shape = self.transpose_shape
        if shape is None:
            return
        self.data = np.transpose(self.data, shape)

    def remap_slow(self, data, mapping=None, write_path=None):
        """
        Reassign volume values (slow on large volumes!) so that they're continuous
        :param data: Volume ndarray
        :param write_path: Where the modified volume will be stored 
            (to spare going through this method next time)
        :param mapping: Pandas Series or a Dictionary that maps raw volume scalars to new ones
        :return: Modified volume data
        """
        logging.info('\nBuilding appropriate volume from Allen data source...')
        #volume = np.vectorize(self.f)(data)
        labels = np.sort(np.unique(data))
        num_labels = len(labels)
        if mapping is None:
            mapping = pd.Series(labels)
        logging.info('Num regions labeled in volume ' + str(num_labels) + ' from ' + str(mapping.size) + ' in atlas')
        logging.info('Reassigning ' + str(num_labels) + ' scalar values...')
        for iter_id in range(num_labels):
            label = labels[iter_id]
            ids = mapping.index[mapping == label].to_list()
            if len(ids) < 1:
                continue
            # On a large volume, this takes a long time
            data[data == label] = ids[0]
            if num_labels > 10000 and iter_id % 10 == 0:
                logging.info('  Progress: ' + str(int(iter_id/num_labels)*100) + '%')
        
        if write_path is not None:
            logging.info('Saving volume data under ' + write_path)
            nrrd.write(write_path, data, index_order='C')
        return data, mapping
        
    def build_lut(self, scalar_map=None, scalar_range=None, color_map=None, 
                    alpha_map=None, zero_is_transparent=True, 
                    noise_amount=0.0, nan_rgba=None, make_active=True):
        """
        Build a look-up table (LUT, sometimes known as transfer function) for the volume
        :param scalar_map: A 2D list with values in first column from the volume itself and values from
            the second column being your scalar values that correspond to such region
        :param scalar_range: Min and max values in a list
        :param color_map: Color map name to apply
        :param alpha_map: Alpha map, either None or a list of values the same length as scalar_map, that
            says how transparent a scalar value should be
        :param zero_is_transparent: Whether zero values are made transparent, True by default
        :param noise_amount: Whether a noise value is applied on the colors
        :param nan_rgba: Color and transparency (RGBA) to assign to invalid (out of range or None) scalar values
        :param make_active: Whether this one is made active (you still have to update the views after that)
        :return: LUTModel
        """
        lut_model = LUTModel()
        lut_model.build(scalar_map, scalar_range, color_map, alpha_map, 
                        zero_is_transparent, noise_amount, nan_rgba)
        self.luts.store(lut_model, set_current=make_active)
        return lut_model


def blend_maps(map1, map2, time, total_time):
    """
    Blend color maps
    """
    weight1 = max(0.0, total_time - time)
    weight2 = max(0.0, time)
    return map1 * weight1 + map2 * weight2


class Volume(vedo.Volume):
    """
    Overwriting of vedo.Volume constructor that is ill-designed as
    it transposes the given numpy array without us knowing about it,
    not giving us the option to choose about that.
    """

    def __init__(self, 
                 inputobj=None,
                 c='RdBu_r',
                 alpha=(0.0, 0.0, 0.2, 0.4, 0.8, 1.0),
                 alphaGradient=None,
                 alphaUnit=1,
                 mode=0,
                 shade=False,
                 spacing=None,
                 dims=None,
                 origin=None,
                 mapper='smart'):

        vtk.vtkVolume.__init__(self)
        vedo.BaseGrid.__init__(self)

        self.axes = [1, 1, 1]

        ###################
        if isinstance(inputobj, str):

            if "https://" in inputobj:
                from vedo.io import download
                inputobj = download(inputobj, verbose=False) # fpath
            elif os.path.isfile(inputobj):
                pass
            else:
                inputobj = sorted(glob.glob(inputobj))

        ###################
        if 'gpu' in mapper:
            self._mapper = vtk.vtkGPUVolumeRayCastMapper()
        elif 'opengl_gpu' in mapper:
            self._mapper = vtk.vtkOpenGLGPUVolumeRayCastMapper()
        elif 'smart' in mapper:
            self._mapper = vtk.vtkSmartVolumeMapper()
        elif 'fixed' in mapper:
            self._mapper = vtk.vtkFixedPointVolumeRayCastMapper()
        elif isinstance(mapper, vtk.vtkMapper):
            self._mapper = mapper
        else:
            print("Error unknown mapper type", [mapper])
            raise RuntimeError()
        self.SetMapper(self._mapper)

        ###################
        inputtype = str(type(inputobj))
        #colors.printc('Volume inputtype', inputtype)

        if inputobj is None:
            img = vtk.vtkImageData()

        elif vedo.utils.isSequence(inputobj):

            if isinstance(inputobj[0], str): # scan sequence of BMP files
                ima = vtk.vtkImageAppend()
                ima.SetAppendAxis(2)
                pb = vedo.utils.ProgressBar(0, len(inputobj))
                for i in pb.range():
                    f = inputobj[i]
                    picr = vtk.vtkBMPReader()
                    picr.SetFileName(f)
                    picr.Update()
                    mgf = vtk.vtkImageMagnitude()
                    mgf.SetInputData(picr.GetOutput())
                    mgf.Update()
                    ima.AddInputData(mgf.GetOutput())
                    pb.print('loading...')
                ima.Update()
                img = ima.GetOutput()

            else:
                if "ndarray" not in inputtype:
                    inputobj = np.array(inputobj)

                if len(inputobj.shape)==1:
                    varr = vedo.numpy2vtk(inputobj, dtype=np.float)
                else:
                    # ------------------------------ Nasty lines commented here
                    #if len(inputobj.shape)>2:
                        #inputobj = np.transpose(inputobj, axes=[2, 1, 0])
                    varr = vedo.numpy2vtk(inputobj.ravel(order='F'), dtype=np.float)
                varr.SetName('input_scalars')

                img = vtk.vtkImageData()
                if dims is not None:
                    img.SetDimensions(dims)
                else:
                    if len(inputobj.shape)==1:
                        vedo.colors.printc("Error: must set dimensions (dims keyword) in Volume.", c='r')
                        raise RuntimeError()
                    img.SetDimensions(inputobj.shape)
                img.GetPointData().SetScalars(varr)

                #to convert rgb to numpy
                #        img_scalar = data.GetPointData().GetScalars()
                #        dims = data.GetDimensions()
                #        n_comp = img_scalar.GetNumberOfComponents()
                #        temp = utils.vtk2numpy(img_scalar)
                #        numpy_data = temp.reshape(dims[1],dims[0],n_comp)
                #        numpy_data = numpy_data.transpose(0,1,2)
                #        numpy_data = np.flipud(numpy_data)

        elif "ImageData" in inputtype:
            img = inputobj

        elif isinstance(inputobj, vedo.Volume):
            img = inputobj.GetMapper().GetInput()

        elif "UniformGrid" in inputtype:
            img = inputobj

        elif hasattr(inputobj, "GetOutput"): # passing vtk object, try extract imagdedata
            if hasattr(inputobj, "Update"):
                inputobj.Update()
            img = inputobj.GetOutput()

        elif isinstance(inputobj, str):
            from vedo.io import loadImageData, download
            if "https://" in inputobj:
                inputobj = download(inputobj, verbose=False)
            img = loadImageData(inputobj)

        else:
            vedo.colors.printc("Volume(): cannot understand input type:\n", inputtype, c='r')
            return

        if dims is not None:
            img.SetDimensions(dims)

        if origin is not None:
            img.SetOrigin(origin) ### DIFFERENT from volume.origin()!

        if spacing is not None:
            img.SetSpacing(spacing)

        self._data = img
        self._mapper.SetInputData(img)
        self.mode(mode).color(c).alpha(alpha).alphaGradient(alphaGradient)
        self.GetProperty().SetShade(True)
        self.GetProperty().SetInterpolationType(1)
        self.GetProperty().SetScalarOpacityUnitDistance(alphaUnit)

        # remember stuff:
        self._mode = mode
        self._color = c
        self._alpha = alpha
        self._alphaGrad = alphaGradient
        self._alphaUnit = alphaUnit


@dataclass
class LUTModel:
    """
    This class might look slightly convoluted but it's actually simple.

    We use double mapping here in order to enable live/interactive visualization
    of volumetric data. Instead of replacing values in a 3D volume, we only replace
    the colors in the 1D LUT list.
    
    The point is that it's too slow to update a given data, like a segmented
    volume with custom values. Instead, we map such custom values to a 1D
    array (our LUT) that maps colors to raw volume values.
    
    This is much faster in terms of rendering and it enables interactive visualization.
    The scalar_lut is the original LUT for the given scalars (custom values)
    and the mapped_lut is the LUT assigned to the surfaces (like slices)
    that have copied data from the volume. The volume is given color_map 
    and alpha_map through vedo methods.

    You might say "ok for double mapping, it's the only way for interactive
    rendering of a volume, but what about color_map and mapped_lut? Aren't
    they the same?". The answer is: they're the same but VTK does not accept
    a vtkLookupTable for a volume. Instead, it wants a vtkColorTransferFunction
    and a vtkPiecewiseFunction for alpha. There's no way around it.
    The color_map will be computed as a vtkColorTransferFunction and
    the alpha_map as the vtkPiecewiseFunction.
    """

    name: str = NotImplementedError

    color_map_function: Any = None
    scalar_map: np.ndarray = None
    scalar_min: float = 0.0
    scalar_max: float = 1.0
    scalar_lut: vtk.vtkLookupTable = None
    mapped_lut: vtk.vtkLookupTable = None
    
    color_map: np.ndarray = None
    alpha_map: np.ndarray = None
    base_color_map: np.ndarray = None


    def build(self, scalar_map=None, scalar_range=None, color_map=None, 
                    alpha_map=None, zero_is_transparent=True, 
                    noise_amount=0.0, nan_rgba=None):
        """
        Build several look-up tables (LUT, sometimes known as transfer function) for the volume.
        This is where double-mapping occurs for segmented volumes that have values from 0 to n where
        each value defines a sub-volume or region. If we want to assign values (say from another model)
        to these regions, we'd have to change the volume values and it would be too slow iterating over
        each voxel in 3D. Instead we define colors that represent these values and assign them to 
        segmented regions in a 1D list.
        :param scalar_map: A 2D list with values in first column from the volume itself and values from
            the second column being your scalar values that correspond to such region
        :param scalar_range: Min and max values in a list
        :param color_map: Color map name to apply
        :param alpha_map: Alpha map, either None or a list of values the same length as scalar_map, that
            says how transparent a scalar value should be
        :param zero_is_transparent: Whether zero values are made transparent, True by default
        :param noise_amount: Whether a noise value is applied on the colors
        :param nan_rgba: Color and alpha values to assign to invalid (out of range or None) scalar values
        :return: LUTModel
        """
        if color_map is None:
            return
        if nan_rgba is None:
            nan_rgba = [0.0, 0.0, 0.0, 0.0]
        if self.base_color_map is None:
            self.base_color_map = color_map
        
        colors = []
        alphas = []
        lut = vtk.vtkLookupTable()
        scalar_lut = vtk.vtkLookupTable()

        # Use the number of values in the volume
        num_steps = len(self.base_color_map) if self.base_color_map is not None else len(color_map)
        num_steps = 2655
        s_min = 0
        s_max = num_steps
        if scalar_map is None:
            if color_map is None and self.base_color_map is not None:
                color_map = self.base_color_map
                
        loop = range(num_steps)
        noise = None
        if isinstance(noise_amount, float) and noise_amount > 0:
            noise = np.random.rand(num_steps) * noise_amount - noise_amount / 2

        # Vedo works with nested lists: 
        # [region_id, [r, g, b]] for color, and [region_id, a] for alpha

        if scalar_map is None:
            # Standard volume that is not segmented
            lut.SetRange(s_min, s_max)
            lut.SetNumberOfTableValues(num_steps)
            scalar_lut.SetRange(s_min, s_max)
            scalar_lut.SetNumberOfTableValues(num_steps)
            for r_id in loop:
                color = vedo.colors.getColor(color_map[r_id])
                color = np.array(color)
                if noise is not None:
                    color = color + noise[r_id]
                    color = np.maximum(color, 0.0)
                    color = np.minimum(color, 1.0)
                colors.append([r_id, color])
                alpha = 1.0 if alpha_map is None else alpha_map[r_id]
                if r_id == 0 and zero_is_transparent:
                    alpha = 0.0
                alphas.append([r_id, alpha])
                lut.SetTableValue(r_id, *color, alpha)
                scalar_lut.SetTableValue(r_id, *color, alpha)
                #scalar_map[r_id] = color_map[r_id]
        else:
            # Segmented volume
            s_min, s_max = scalar_range
            lut.SetRange(0, num_steps)
            lut.SetNumberOfTableValues(num_steps)
            color = None
            for r_id in range(num_steps):
                try:
                    value = scalar_map[r_id]
                except Exception:
                    value = None
                if value is None:# or s_min > value or s_max < value:
                    color = nan_rgba[:3]
                    alpha = nan_rgba[3]
                else:
                    color = vedo.colorMap(value, color_map, s_min, s_max)
                    alpha = 1.0 if alpha_map is None else alpha_map[r_id]
                    if value == 0 and zero_is_transparent:
                        alpha = 0.0
                colors.append([r_id, color])
                alphas.append([r_id, alpha])
                lut.SetTableValue(r_id, *color, alpha)
                
            # Real scalar LUT, mainly as a reference for the user
            # Here the colors resulting from the given scalar min to max
            # are assigned to segmented values in the volume
            mock_values = np.linspace(s_min, s_max, num_steps)
            scalar_lut.SetRange(s_min, s_max)
            scalar_lut.SetNumberOfTableValues(len(mock_values))
            for r_id in range(len(mock_values)):
                color = list(vedo.colorMap(mock_values[r_id], color_map, s_min, s_max))
                alpha = 0.0 if mock_values[r_id] == 0 and zero_is_transparent else 1.0
                scalar_lut.SetTableValue(r_id, *color, 1.0)

        lut.Build()
        scalar_lut.Build()

        # Just to avoid confusion: the user can give a string as a color map, like 'viridis'
        # but the real color map object is stored in self.color_map. The name 'viridis'
        # is stored under self.color_map_function (if needed later on)
        self.color_map_function = color_map
        self.color_map = colors
        self.alpha_map = alphas
        self.scalar_map = scalar_map
        self.mapped_lut = lut
        self.scalar_lut = scalar_lut

    def get_sorted_scalars(self):
        """
        Get a numpy 2D array of key-value pairs sorted by value
        :return: 2D array
        """
        sorted_scalars = np.zeros((len(self.scalar_map), 2))
        values = list(self.scalar_map.values())
        keys = list(self.scalar_map.keys())
        sorted_scalars[:, 0] = keys
        sorted_scalars[:, 1] = values
        sorted_mask = sorted_scalars[:, 1].argsort()
        sorted_scalars = sorted_scalars[sorted_mask]
        return sorted_scalars



class VolumeController():
    """
    Wrapper class that handles both the volume and its slices
    """

    def __init__(self, plot, model, initialize=True, clipping=True, slicer_box=True, 
                center_on_edges=False, alpha_unit_upper_offset=0.0, add_to_scene=True):
        """
        Constructor
        :param plot: Plot instance
        :param model: VolumeModel instance
        :param initialize: Whether the initalization
        :param clipping: Whether clipping is enabled at init time
        :param slicer_box: Whether the slicer box is enabled at init
        :param center_on_edges: Whether the volume is offest by half a voxel or not
        :param alpha_unit_upper_offset: The offset to apply to alpha unit computation.
            If greater than 0, the volume will be less opaque
        :param add_to_scene: Whether the volume is added to scene after init
        """
        self.plot = plot
        self.model = model

        self.actor = None
        self.picker = None
        self.scalars = None
        self.mask = None
        self.bounding_mesh = None
        self.alpha_unit_upper_offset = alpha_unit_upper_offset
        self.alpha_factor = 0.001 # * self.model.resolution

        self.clipping_planes = None
        self.enable_volume_clipping = True
        self.clipping_axes = []
        self.slicers = OrderedDict()
        self.slicers_selectable = False
        self.scalar_bar = None
        
        if initialize:
            self.initialize(clipping, slicer_box, center_on_edges, add_to_scene)
        #msg = 'Volume abs center', self.volume_center, 'position', np.array(self.volume_actor.pos())
        #logging.info(msg)

    def get_related_actors(self):
        """
        Get all 3D actors related to this view (for registering it in the application)
        :return: List of VTK objects
        """
        actors = []
        for slicer_id in self.slicers:
            actor = self.slicers[slicer_id].actor
            if actor is not None:
                actors.append(actor)
        for iso_id in self.model.isosurfaces:
            actors.append(self.model.isosurfaces[iso_id])
        actors.append(self.actor)
        return actors

    def initialize(self, clipping=True, slicer_box=True, center_on_edges=False, add_to_scene=True):
        """
        Set the volume actor for visualization in VTK
        :param clipping: Whether clipping is enabled
        :param slicer_box: Whether the slicer box mode is enabled (6 clipping planes)
        :param center_on_edges: Whether the volume's center is aligned to its edges 
            rather than the voxel center
        :param add_to_scene: Whether the object is added to the scene
        """
        self.build_actor(center_on_edges, add_to_scene)
        self.initialize_picker()

        if slicer_box:
            self.initialize_slicer_box()
        self.initialize_clipping_planes()
        self.set_volume_clipping(clipping)
        self.set_color_map()

        '''
        if use_mask:
            self.mask = self.actor.clone()
            self.mask.threshold(1, replace=1, replaceOut=0)
            self.actor.mapper().SetMaskTypeToBinary()
            self.actor.mapper().SetMaskInput(self.mask)
        '''

    def set_volume_visibility(self, on=True):
        """
        Set volume visibility
        :param on: Visibility boolean
        """
        if self.actor is not None:
            self.actor.SetVisibility(on)

    def set_slices_visibility(self, on=True):
        """
        Set the visibility of slices
        :param on: Visibility boolean
        """
        for slicer_id in self.slicers:
            slicer_view = self.slicers.get(slicer_id)
            slicer_view.actor.SetVisibility(on)

    def get_slices_opacity(self):
        """
        Get the opacity of slices (should be the same value for all slices)
        A mean calculation is performed on all slices alpha, just in case
        :return: Alpha value
        """
        value = 0
        num_values = 0
        for slicer_id in self.slicers:
            slicer = self.slicers[slicer_id]
            if slicer.actor is not None:
                slice_alpha = slicer.actor.GetProperty().GetOpacity()
                if slice_alpha is None:
                    continue
                value += slice_alpha
                num_values += 1
        if num_values == 0 or value == 0:
            return None
        return value / num_values

    def set_slices_opacity(self, value):
        """
        Set the opacity of slices
        :param value: Alpha value
        """
        for slicer_id in self.slicers:
            slicer = self.slicers[slicer_id]
            if slicer.actor is not None:
                slicer.actor.alpha(value)

    def get_opacity(self):
        """
        Get the relative opacity unit
        :return: Float
        """
        return self.get_relative_opacity_unit()

    def get_relative_opacity_unit(self):
        """
        Get the alpha unit relative value
        :return: Float
        """
        alpha_unit = self.actor.alphaUnit()
        r = self.model.resolution
        # Inverse function of set_opacity_unit()
        value = 1.1 - (alpha_unit / r)**0.5
        return value

    def set_opacity(self, value):
        """
        Set the opacity of the volume like in set_opacity_unit()
        :param value: Opacity value between 0.0 and 1.0
        :return: Resulting alpha unit
        """
        self.set_opacity_unit(value)

    def set_opacity_unit(self, value):
        """
        Set the opacity of the volume by modifying its alpha unit (a VTK thing).
        The alpha unit defines how much a voxel is transparent to incoming ray.
        This method normalizes the range between 0.0 and 1.0 as it depends
        on the resolution of the volume
        :param value: Opacity value between 0.0 and 1.0
        :return: Resulting alpha unit
        """
        r = self.model.resolution
        # 1 is chosen and not 1.0 because when value == 1.0, that would
        # mean that the volume is fully opaque and this yields artifacts with VTK
        alpha_unit = (1 + self.alpha_unit_upper_offset - value)**2 * r
        # vedo calls it "alpha" unit, vtk "opacity" unit. same-same!
        self.actor.alphaUnit(alpha_unit)
        return alpha_unit

    def get_spacing(self):
        """
        Get the spacing/resolution of the volume
        """
        res = self.model.resolution
        spacing = None
        if isinstance(res, int) or isinstance(res, float):
            spacing = np.array([res]*3)
        elif len(res) == 3:
            spacing = res
        else:
            raise ValueError(f'Given volume resolution {self.model.resolution} is invalid')
        return spacing

    def build_actor(self, center_on_edges=False, add_to_scene=True): #[1, 2]
        """
        Set the volume actor for visualization in VTK
        :param center_on_edges: Whether alignment by one voxel is applied
        :param add_to_scene: Whether the object is added to the scene
        """
        spacing = self.get_spacing()
        self.actor = Volume(self.model.data, spacing=spacing, mapper='smart')
        self.scalars = self.actor._data.GetPointData().GetScalars()
        self.actor.name = self.model.name
        self.actor.shade(False)
        self.actor.mode(0)
        self.actor.pickable(True)
        self.set_interactive_subsampling(False)

        if center_on_edges:
            # Moving the volume by one voxel. This is possibly due the use of custom spacing.
            self.actor.pos(self.actor.pos() + spacing)
            center = np.array(self.actor.pos()) + self.actor.center()
            if np.linalg.norm(center - self.model.center) > 0:
                #print('Adjusting volume center from', self.model.center, 'to', center)
                self.model.center = center
        
        self.set_opacity_unit(0.9)
        self.actor.jittering(True)
        #self.actor._mapper.AutoAdjustSampleDistancesOn()
        #self.actor._mapper.SetBlendModeToAverageIntensity()
        #self.actor._mapper.SetSampleDistance(100)

        if add_to_scene:
            self.plot.add(self.actor, render=False)

    def set_position(self, position):
        """
        Set the position of the volume
        """
        self.actor.pos(position)
        # TODO: we're entering in unstable things when we move the volume
        # because there is not yet a guaranteed support for updating the slices 
        # with the correct position
        self.reset_clipping_planes()

    def mirror_volume(self, axes):
        """
        Mirror the volume on given axes
        :param mirror_axes: A list of axes (either 0, 1, 2 or 'x', 'y', 'z') on which
            the volume will be mirrored. Optional
        """
        if axes is None or self.actor is None:
            return
        axes_str = ['x', 'y', 'z']
        for axis in axes:
            if isinstance(axis, int) and 0 <= axis <= 2:
                axis = axes_str[axis]
            if isinstance(axis, str) and len(axis) == 1:
                self.actor.mirror(axis=axis.lower())

    def initialize_picker(self, opacity_iso_value=0.0001):
        """
        Initialize the volume picker
        :param opacity_iso_value: Threshold that defines at what accumulated
            opacity the picker hits the volume. In the case of a segmented volume,
            you want to keep this value very low as the default one.
        """
        # As per C++ doc https://vtk.org/Wiki/VTK/Examples/Cxx/VTKConcepts/Scalars
        # https://stackoverflow.com/questions/35378796/vtk-value-at-x-y-z-point 
        picker = vtk.vtkVolumePicker()
        picker.PickCroppingPlanesOn()
        picker.UseVolumeGradientOpacityOff()
        picker.SetTolerance(opacity_iso_value)
        # A low OpacityIsoValue is necessary in the case of segmented volumes
        picker.SetVolumeOpacityIsovalue(opacity_iso_value)
        picker.AddPickList(self.actor)
        picker.PickFromListOn()
        self.picker = picker
    
    def initialize_slicer_box(self):
        """
        Initialize 6 slicing planes as a box.
        """
        for axis_id in range(6):
            slicer_model = SlicerModel(axis=axis_id)
            slicer_model.align_to_axis(axis_id, self.model.dimensions)
            self.model.slicers.store(slicer_model)
            # It's important in this case to have standalone=False
            self.slicers[axis_id] = SlicerView(self.plot, self, slicer_model, standalone=False)
        
    def update_slicer(self, slicer_id, value=None, normal=None):
        """
        Update a given slicer with the given value
        :param slicer_id: SlicerView id
        :param value: Value or 3D point
        :param normal: Normal
        """
        slicer_view = self.slicers.get(slicer_id)
        if slicer_view is None:
            return

        # This is an important part where the slicing plane is itself sliced by other planes
        slicer_model = slicer_view.model
        slicer_model.clipping_planes = self.get_clipping_planes(slicer_model.axis)
        
        # Use given value (or point) and normal to guide the below code
        result = slicer_model.update(value, normal)
        if not result:
            return

        # Update slicing image
        slicer_view.update()

    def initialize_clipping_planes(self):
        """
        Initialize X, Y and Z clipping planes with two planes per axis 
        for positive and negative slicing
        """
        self.clipping_planes = vtk.vtkPlaneCollection()
        slicer_models = self.model.slicers
        for slicer_id in slicer_models:
            self.clipping_planes.AddItem(vtk.vtkPlane())
        self.reset_clipping_planes()
        return

    def get_clipping_planes(self, except_axis=None):
        """
        Get the current clipping planes except the ones on the given axis
        :param except_axis: Axis id to ignore. If None, all clipping planes will be returned
        :return: vtkPlaneCollection
        """
        if not isinstance(except_axis, int):
            return self.clipping_planes
        exceptions = [except_axis * 2, except_axis * 2 + 1]
        planes = vtk.vtkPlaneCollection()
        for plane_id in range(self.clipping_planes.GetNumberOfItems()):
            if plane_id in exceptions:
                continue
            plane = self.clipping_planes.GetItem(plane_id)
            planes.AddItem(plane)
        return planes

    def reset_clipping_planes(self):
        """
        Reset clipping planes
        """
        slicer_models = self.model.slicers
        for slicer_id in slicer_models:
            slicer_model = slicer_models[slicer_id]
            plane_id = slicer_model.get_box_plane_id()
            plane = self.clipping_planes.GetItem(plane_id)
            plane.SetOrigin(slicer_model.origin + self.actor.pos())
            plane.SetNormal(slicer_model.normal)

    def clip_on_axis(self, position=None, axis=None, normal=None):
        """
        Apply clipping on a single axis
        :param position: Position
        :param axis: Clipping axis, defaults to 0 (X axis)
        :param thickness: Whether a thickness (so two clipping planes) are applied
        """
        axis_offset = 0
        # This should already be sorted in the model but in case it isn't, we double check here
        if normal is not None and normal[axis] < 0:
            # This means that the given axis has two 
            # clipping planes and we take the negative one
            axis_offset += 1
            #position = self.model.dimensions - position
        axis_storage_id = axis * 2 + axis_offset
        plane = self.clipping_planes.GetItem(axis_storage_id)
        plane.SetOrigin(position)
        plane.SetNormal(normal)

    def set_volume_clipping(self, on=None):
        """
        Set volume clipping on or off.
        :param on: Whether clipping is enabled or disabled. If None, then
        the state is toggled.
        """
        if on is None:
            self.enable_volume_clipping = not self.enable_volume_clipping
        else:
            self.enable_volume_clipping = on
        if self.enable_volume_clipping:
            self.actor.mapper().SetClippingPlanes(self.clipping_planes)
        else:
            self.actor.mapper().SetClippingPlanes(None)

    def clip_to_bounds(self, bounds):
        """
        Clip the volume and move the slicing planes according to 6 boundary points
        :param bounds: Six values in a list (xmin, xmax, ymin, ymax, zmin, zmax)
        """
        planes = vtk.vtkPlanes()
        planes.SetBounds(bounds)
        # Normals are reversed with the above code
        # so we fix that here with flip_normals=True
        self.set_clipping_planes(planes, flip_normals=True)

    def box_widget_update(self, widget=None, event=None):
        """
        Clip the volume with the current box widget
        :param widget: vtkBoxCutter
        :param event: vtkEvent
        """
        if widget is None:
            return
        planes = vtk.vtkPlanes()
        widget.GetPlanes(planes)
        self.set_clipping_planes(planes)

    def set_clipping_planes(self, planes, flip_normals=False):
        """
        Clip the volume and move the slicing planes according the given planes
        :param planes: vtkPlanes
        """
        vtk_n = planes.GetNormals()
        vtk_pts = planes.GetPoints()
        num_pts = vtk_pts.GetNumberOfPoints()
        for plane_id in range(num_pts):
            normal = vtk_n.GetTuple(plane_id)
            origin = vtk_pts.GetPoint(plane_id)
            plane = self.clipping_planes.GetItem(plane_id)
            current_origin = np.array(plane.GetOrigin())
            # We don't need to check the normal because
            # we prevent box cutter rotation in our case
            if np.linalg.norm(current_origin - origin) < 0.1:
                continue
            plane.SetOrigin(origin)
            if flip_normals:
                normal = np.array(normal)*-1
            plane.SetNormal(normal)
            self.update_slicer(plane_id, origin, normal)
        self.clipping_planes.Modified()
        self.actor.GetMapper().Update()

    def set_alpha_map(self, alpha_map, alpha_factor=None):
        """
        Set alpha map to the volume view
        :param alpha_map: 2D list of scalar values and alpha values
        :param alpha_factor: Alpha factor
        """
        if alpha_map is None:
            if self.model.luts.current is None:
                return
            alpha_map = self.model.luts.current.alpha_map
        if alpha_factor is None:
            alpha_factor = self.alpha_factor
        if len(np.array(alpha_map).shape) > 1:
            volume_alpha_map = np.ones_like(alpha_map).astype(float)
            volume_alpha_map[:] = alpha_map[:]
            volume_alpha_map[:, 1] *= alpha_factor
            self.actor.alpha(volume_alpha_map)
        else:
            self.actor.alpha(np.array(alpha_map) * alpha_factor)

    def set_color_map(self, color_map=None, alpha_map=None):
        """
        Set the color and alpha map to the view objects
        :param color_map: Nested list of scalar values and rgb colors
            like [[0, [0.0, 0.0, 0.0]], [8, [0.5, 0.8, 0.3]], ...]
        :param alpha_map: 2D list of scalar values and alpha values
        """
        lut = self.model.luts.current
        if color_map is None and lut is not None:
            color_map = lut.color_map
        if alpha_map is None and lut is not None:
            alpha_map = lut.alpha_map
        if color_map is None:
            return
        self.actor.cmap(color_map)
        self.set_alpha_map(alpha_map)
        
        if lut is not None:
            for surface in self.model.isosurfaces:
                surface._mapper.SetLookupTable(lut.opaque_lut)
            for slicer_id in self.slicers:
                slicer = self.slicers[slicer_id]
                slicer.apply_lut(lut.mapped_lut)
        else:
            for slicer_id in self.slicers:
                slicer = self.slicers[slicer_id]
                slicer.set_color_map(color_map, alpha_map)

    def disable_shading(self):
        """
        Disable volume shading
        """
        volumeProperty = self.actor.GetProperty()
        volumeProperty.ShadeOff()
        self.actor.SetProperty(volumeProperty)

    def enable_shading(self, ambient=0.6, diffuse=0.8, specular=0.9):
        """
        Enable volume shading
        TODO: See if this method is useful
        """
        volumeProperty = self.actor.GetProperty()
        volumeProperty.SetInterpolationTypeToLinear()
        volumeProperty.ShadeOn()
        volumeProperty.SetAmbient(ambient)
        volumeProperty.SetDiffuse(diffuse)
        volumeProperty.SetSpecular(specular)
        volumeProperty.SetScalarOpacityUnitDistance(1)
        self.actor.SetProperty(volumeProperty)

    def toggle_slices_visibility(self):
        """
        Toggle slices visibility
        """
        self.model.slices_visible = not self.model.slices_visible
        for slicer_id in self.slicers:
            slicer = self.slicers[slicer_id]
            self.update_slicer(slicer)
            if slicer.actor is not None:
                slicer.actor.SetVisibility(self.model.slices_visible)

    def toggle_hollow(self):
        """
        Toggle hollow mode for volume rendering. This is intended 
        to work only on segmented (annotated) volumes.
        """
        volume_property = self.actor.GetProperty()
        # Shout at VTK devs: it's twisted to name properties Disable and then have DisableOff...
        disabled = bool(volume_property.GetDisableGradientOpacity())
        if disabled:
            volume_property.DisableGradientOpacityOff()
            alpha_gradient = vtk.vtkPiecewiseFunction()
            alpha_gradient.AddPoint(0, 0.0)
            alpha_gradient.AddPoint(1, 0.75)
            alpha_gradient.AddPoint(2, 1.0)
            volume_property.SetGradientOpacity(alpha_gradient)
        else:
            volume_property.DisableGradientOpacityOn()
        return not disabled

    def get_value_from_xyz(self, position, normal_step=None, avoid_values=0, cast_to_int=True, none_as_zero=False):
        """
        Get a scalar value from the volume with respect to XYZ coordinates and a optionally a normal step,
        that is the normal on which to probe multiplied by the distance you want to travel further into
        the volume to pick a correct value. Often the "surface point" on a volume with non uniform transparency
        is at the boundary between transparent (let's say a 0 value is transparent) and more opaque parts.
        So you need to go further into the "cloud" so to speak, in order to find the values you want.
        :param position: 3D array
        :param normal_step: A vector normal multiplied by the lookup distance, in case the raw position yields
            bad or unwanted results
        :param avoid_values: Try and find other values than this
        :param cast_to_int: Whether the value should be cast to integer
        :return: Scalar value
        """
        if isinstance(avoid_values, int) or isinstance(avoid_values, float):
            avoid_values = [avoid_values]
        # TODO: see if this is faster? To be tested
        # ijk_result = [0.0, 0.0, 0.0]
        # volume_actor._data.TransformPhysicalPointToContinuousIndex(xyz, ijk_result)
        # volume_actor._data.GetPoint(ijk_result)
        pt_id = self.actor._data.FindPoint(*position)
        valid_id = 0 < pt_id < self.scalars.GetNumberOfValues()
        value = self.scalars.GetValue(pt_id) if valid_id else None
        if not valid_id or (value in avoid_values):
            if normal_step is not None:
                position += normal_step
                pt_id = self.actor._data.FindPoint(*position)
                valid_id = 0 < pt_id < self.scalars.GetNumberOfValues()
                value = self.scalars.GetValue(pt_id) if valid_id else None
        if cast_to_int and value is not None:
            value = int(value)
        if value is None and none_as_zero:
            value = 0
        return value

    def raycast(self, origin, screen_position):
        """
        Shorthand for pick() method
        """
        return self.pick(origin, screen_position)

    def pick(self, origin, screen_position):
        """
        Find the nearest intersection – even on sliced volume – with the ray formed
        by an origin and a screen-space position (given by VTK when you click on an actor)
        :param origin: Origin of the vector
        :param screen_position: 2D position on screen. This is given by vtk events like MouseRelease
        :return: The nearest position and its related value queried in the volume image
        """
        self.picker.Pick(*screen_position[:2], 0, self.plot.renderer)
        position = np.array(self.picker.GetPickPosition())
        ray = position - origin
        distance = np.linalg.norm(ray)
        normal = ray / distance

        # Go half a voxel further to make sure we don't hit "void"
        vol_position = position # + normal * self.model.resolution / 2
        probe_position = position + normal * self.model.resolution * 10

        closest_dist = distance
        slice_position = None
        # See if the line hits any of the slicers (that are image planes)
        for slicer_id in self.slicers:
            slicer = self.slicers[slicer_id]
            if slicer.got_slice:
                hits = slicer.actor.intersectWithLine(origin, probe_position)
                if len(hits) != 1:
                    continue
                new_dist = np.linalg.norm(position - hits[0])
                if new_dist < closest_dist and new_dist < self.model.resolution * 2:
                    closest_dist = new_dist
                    slice_position = hits[0]

        if slice_position is None:
            position = vol_position
        else:
            position = slice_position
        value = self.get_value_from_xyz(position, normal * self.model.resolution * 4)
        return position, value

    def add_probe(self, origin, destination, resolution=40, radius=10, color_map=None, 
                screen_space=True, min_v=None, max_v=None, add_to_scene=True):
        """
        Add a series of points along a line probe
        :param origin: Probe origin
        :param destination: Probe destination point
        :param resolution: Number of (equidistant) points that will be probed along that line
        :param radius: Radius of the points
        :param color_map: Scalars color map
        :param screen_space: Whether the points are screen space or spheres
        :param min_v: Min scalar value
        :param max_v: Max scalar value
        :param add_to_scene: Whether the new probe is added to scene
        :return: Points
        """
        if color_map is None:
            color_map = self.model.luts.current.color_map
        
        positions, values = self.probe(origin, destination, resolution)

        points_obj = obj.Points(positions, values=values, radius=radius, screen_space=screen_space,
                                color_map=color_map, min_v=min_v, max_v=max_v)
        
        points_obj.origin = origin
        points_obj.destination = destination
        # Dynamic properties assignment
        points_obj.target = self.actor
        points_obj.target_controller = self

        if add_to_scene:
            self.plot.add(points_obj)
        return points_obj

    def update_probe(self, origin, destination, points_obj):
        """
        Update a probe with given start and end points
        :param origin: Start point
        :param destination: End point
        :param points_obj: Points object
        """
        resolution = points_obj._polydata.GetPoints().GetNumberOfPoints()
        positions, values = self.probe(origin, destination, resolution)
        points_obj.update_data(positions, values)

    def probe(self, origin, destination, resolution=40):
        """
        Probe a volume with a line
        :param origin: Origin of the line probe
        :param destination: Destination of the line probe
        :param resolution: Number of point samples along the probe
        :return: Positions and values
        """
        origin = np.array(origin)
        destination = np.array(destination)
        distance = np.linalg.norm(destination - origin)
        ray = destination - origin
        ray_norm = ray / distance
        step = distance / resolution
        positions = [origin + ray_norm * p_id * step for p_id in range(resolution)]
        values = np.array([self.get_value_from_xyz(point, none_as_zero=True) for point in positions])
        return positions, values

    def set_interactive_subsampling(self, on=False):
        """
        Set volume subsampling on or off. 
        This is enabled by default in VTK and we disable it by default in IBLViewer
        :param on: Whether volume subsampling in interactive mode is on or off
        """
        #self.plot.window.SetDesiredUpdateRate(0)
        #self.actor._mapper.SetInteractiveUpdateRate(0)
        self.model.interactive_subsampling = on
        self.actor._mapper.SetAutoAdjustSampleDistances(on)
        if on:
            self.actor._mapper.InteractiveAdjustSampleDistancesOn()
        else:
            self.actor._mapper.InteractiveAdjustSampleDistancesOff()

    def isosurface(self, label, exceptions=[0], force_rebuild=False, set_current=True, to_int=True, split_meshes=True):
        """
        Creates a surface mesh (isosurface) of a segmented/labelled volume for the given value.
        Unlike general isosurfacing, this method extracts only the surface mesh of the 
        desired region/label/segmentation, not of all values from 0 to label.
        :param label: Label (scalar) value found in the volume
        :param exceptions: If the label is found in the exceptions list, isosurfacing will not occur
        :param force_rebuild: Whether rebuilding is forced in case we find an existing mesh for the given label
        :param set_current: Whether the label is set as the current one in the model
        :param to_int: Whether the label is cast to integer
        :param split_meshes: Whether we split meshes when multiple ones are found
        :return: A list of all manifold meshes for the given label
        """
        if label is None or label in exceptions:
            return
        if to_int:
            label = int(label)
        existing_meshes = self.model.isosurfaces.get(label)
        if existing_meshes is not None and not force_rebuild:
            return existing_meshes

        lut = self.model.luts.current
        simple_lut = vtk.vtkLookupTable()
        simple_lut.SetNumberOfColors(1)
        simple_lut.SetTableRange(0, 1)
        simple_lut.SetScaleToLinear()
        simple_lut.SetTableValue(0, 0, 0, 0, 0)
        simple_lut.SetTableValue(1, *lut.mapped_lut.GetTableValue(label))
        simple_lut.Build()

        # Generate object boundaries from labelled volume 
        discrete = vtk.vtkDiscreteMarchingCubes()
        discrete.SetInputData(self.actor.imagedata())
        discrete.GenerateValues(1, label, label)

        smoothing_iterations = 15
        pass_band = 0.001
        feature_angle = 120.0

        smoother = vtk.vtkWindowedSincPolyDataFilter()
        smoother.SetInputConnection(discrete.GetOutputPort())
        smoother.SetNumberOfIterations(smoothing_iterations)
        smoother.BoundarySmoothingOff()
        smoother.FeatureEdgeSmoothingOff()
        smoother.SetFeatureAngle(feature_angle)
        smoother.SetPassBand(pass_band)
        smoother.NonManifoldSmoothingOn()
        smoother.NormalizeCoordinatesOn()
        smoother.Update()
        
        self.model.isosurfaces[label] = []
        #splitter = vtk.vtkExtractPolyDataGeometry()
        if split_meshes:
            splitter = vtk.vtkPolyDataConnectivityFilter()
            splitter.SetInputConnection(smoother.GetOutputPort())
            splitter.SetExtractionModeToAllRegions()
            splitter.ColorRegionsOn()
            splitter.Update()

            for region_id in range(splitter.GetNumberOfExtractedRegions()):
                #splitter.AddSpecifiedRegion(region_id)
                #splitter.Update()
                
                #poly = vtk.vtkPolyData()
                #poly.ShallowCopy(splitter.GetOutput())

                threshold = vtk.vtkThreshold()
                threshold.SetInputConnection(splitter.GetOutputPort())
                threshold.ThresholdBetween(region_id, region_id)
                threshold.Update()
                actor = vedo.Mesh(threshold.GetOutput())
                #actor._mapper.SetScalarRange(min_value, lut.scalar_max)
                #actor._mapper.SetUseLookupTableScalarRange(True)
                actor._mapper.SetLookupTable(simple_lut)
                actor._mapper.ScalarVisibilityOn()
                actor.name = 'Isosurface_' + str(label)
                self.model.isosurfaces[label].append(actor)
                #actor.cmap(lut.scalar_lut, np.ones(poly.GetNumberOfVerts())*label)
        else:
            poly = smoother.GetOutput()
            actor = vedo.Mesh(poly)
            actor._mapper.SetLookupTable(simple_lut)
            actor._mapper.ScalarVisibilityOn()
            actor.name = 'Isosurface_' + str(label)
            self.model.isosurfaces[label].append(actor)

        '''
        pdnorm = vtk.vtkPolyDataNormals()
        pdnorm.SetInputData(smoother.GetOutput())
        pdnorm.ComputePointNormalsOn()
        pdnorm.ComputeCellNormalsOn()
        pdnorm.FlipNormalsOff()
        pdnorm.ConsistencyOn()
        pdnorm.Update()
        
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(smoother.GetOutputPort())
        mapper.SetLookupTable(lut.scalar_lut)
        mapper.SetScalarRange(min_value, lut.scalar_max)
        '''
        if set_current:
            self.model.isosurfaces.set_current(label)
        return self.model.isosurfaces[label]


@dataclass
class SlicerModel:
    PREFIX = '[Slicer]_'
    MIN_SLAB_THICKNESS = 1.0 #um

    __count = 0

    def unique_name():
        SlicerModel.__count += 1
        return f'{SlicerModel.PREFIX}_{SlicerModel.__count}'

    name: str = field(default_factory=unique_name)
    # 0, 1 or 2. See the normal for axis orientation
    axis: int = None
    value: float = 0.0
    bounds: np.ndarray = None
    #thickness: float = 0.0
    origin: np.ndarray = np.array([0.0, 0.0, 0.0])
    normal: np.ndarray = np.array([1.0, 0.0, 0.0])
    clipping_planes: vtk.vtkPlaneCollection = None

    def get_box_plane_id(self):
        """
        Get the plane id
        :return: Int
        """
        if self.axis is None:
            return
        offset = 0 if self.normal[self.axis] < 0 else 1
        return self.axis * 2 + offset

    def get_axis_aligned_info(self, vtk_axis):
        """
        VTK stores box clipping planes in the order:
        -X to +X: 0, 1
        -Y to +Y: 2, 3
        -Z to +Z: 4, 5
        This method retrieves what is the XYZ axis (0, 1 or 2)
        and its orientation sign
        :return: Int axis and float orientation
        """
        orientation = -1.0 if vtk_axis % 2 == 0 else 1.0
        axis = (vtk_axis - vtk_axis % 2) // 2
        return axis, orientation

    def align_to_axis(self, axis, dimensions=None):
        """
        Set the axis of the slicer
        :param axis: See parameter vtk_axis in SlicerModel.get_axis_aligned_info()
        :param dimensions: Dimensions of the volume
        """
        if not isinstance(axis, int):
            return
        normal = np.zeros(3).astype(float)
        xyz_axis, orientation = self.get_axis_aligned_info(axis)
        normal[xyz_axis] = orientation
        self.axis = xyz_axis
        if dimensions is not None and orientation < 0:
            self.origin = np.zeros(3)
            self.origin[xyz_axis] = dimensions[xyz_axis]
        self.normal = normal

    def flip_normal(self):
        """
        Flip the normal of the slicer
        """
        self.normal *= -1.0
        self.check_normal()
        if isinstance(self.axis, int):
            self.axis *= -1

    def check_normal(self):
        """
        Check if the normal is axis-aligned.
        If not, the axis is set to None.
        """
        zeros = self.normal == 0
        if len(self.normal[zeros]) >= 2:
            self.axis = 0

    def update(self, value=None, normal=None, axis=None):
        """
        Update slicer
        :param value: Origin of the slicing plane
        :param normal: Normal of the slicing plane
        :param axis: Axis, if the plane is axis-aligned
        :return: True if model changed, False if it didn't
        """
        if not(isinstance(value, int) or isinstance(value, float)):
            if normal is None:
                normal = self.normal
            if normal is None:
                return False
            if normal[1] == 0 and normal[2] == 0:
                axis = 0 #if normal[0] > 0 else 1
            elif normal[0] == 0 and normal[2] == 0:
                axis = 1 #if normal[1] > 0 else 1
            elif normal[0] == 0 and normal[1] == 0:
                axis = 2 #if normal[2] > 0 else 1
            if axis is not None:
                value = value[axis]
        if axis is None:
            axis = self.axis
        if self.value == value:
            return False
        if axis is not None:
            self.value = value
            self.origin = np.array(normal) * value
        else:
            self.value = None
            self.origin = value
        self.normal = normal
        self.axis = axis
        return True


class SlicerView():

    slices = {}

    def __init__(self, plot, volume_view, slicer_model, standalone=True):
        """
        Constructor
        :param plot: Plot instance
        :param volume_view: VolumeView instance
        :param slicer_model: SlicerModel instance
        :param standalone: Whether the slice is a standalone actor that
        can be clicked. Set this to False if you want to use transparency,
        at the expense that because of a VTK bug, you won't be able to
        click on it anymore, requiring you to code another way of detecting
        where the user clicked. See more in initialize_mapper()
        """
        self.plot = plot
        self.volume_view = volume_view
        self.model = slicer_model
        self.actor = None
        self.filter = None
        self.filter = None
        self.actor = None
        self.reslice = None
        self.slice_type = -1
        self.depth_peeling_enabled = None
        self.standalone = standalone
        self.got_slice = False
        self.color_map = None
        self.alpha_map = None
        self.initialize()

    def initialize(self, render=False):
        """
        Initialize the slicer object
        """
        if self.filter is None:
            self.filter = vtk.vtkImageDataGeometryFilter()
        
        if self.actor is None:
            self.actor = vedo.Mesh(self.filter.GetOutput())
            # Adding empty actor so that it's updated later on
            self.plot.add(self.actor, render=render)
            self.actor.lighting('off')
            self.actor.name = self.model.name

        self.initialize_mapper()

    def initialize_mapper(self):
        """
        Initialize the object mapper
        """
        mapper = self.actor._mapper
        mapper.SetScalarModeToUsePointData() #SetScalarModeToUsePointFieldData
        mapper.SetColorModeToMapScalars()
        mapper.ScalarVisibilityOn()
        # We operate on static volumes thanks to the double LUT mapping implemented here
        mapper.SetStatic(True)

        # Without using scalar range, the mapping will be off
        mapper.SetUseLookupTableScalarRange(True)
        
        # We prevent this actor from being pickable as a result of the bug described below
        # when we want to use transparency on the slice.
        self.actor.pickable(self.standalone)
        if self.standalone:
            # There is a bug in VTK 9 that prevents clicking on transparent objects
            # as reported on vedo's tracker https://github.com/marcomusy/vedo/issues/291
            # The "Force opaque fix" below should be gone with the next VTK update hopefully.
            # In the meantime, we use this.
            # TODO: remove this when this bug is fixed in VTK
            self.actor.ForceOpaqueOn()
        else:
            # We bypass the transparent selection bug when a VolumeView has multiple slicers 
            # like in box mode because the click detection occurs on the volume and we perform 
            # an additional test to see if a slicer yields a nearby result. If it does, 
            # the result is like clicking on the slice and we get transparency for free.
            pass

        # Make sure we have depth peeling activated, otherwise transparency with volumes
        # will look weird and in the wrong order
        self.plot.renderer.UseDepthPeelingOn()
        self.plot.renderer.UseDepthPeelingForVolumesOn()
        
        segmented = self.volume_view.model.is_segmented()
        if segmented:
            # This very line below will mess up the entire slice coloring if:
            # - you have a segmented volume and this is set to True
            # - you have a non-segmented (like raw MRI, CT) volume and this is set to False
            mapper.SetInterpolateScalarsBeforeMapping(not segmented)
        mapper.Update()

    def set_color_map(self, color_map, alpha_map=None):
        """
        Set a color map to the slice
        :param color_map: Color map, can be a string, a list of colors or more. 
            See vedo documentation.
        """
        self.color_map = color_map
        if alpha_map is not None:
            self.alpha_map = alpha_map
        if self.got_slice and color_map is not None:
            self.actor.cmap(self.color_map, alpha=self.alpha_map)

    def set_slice_type(self, slice_type):
        """
        Set the slice type. 0 for axial, 1 for free slicing
        :param slice_type: Int value
        """
        if slice_type == 0 and self.slice_type != slice_type:
            self.slice_type = slice_type
            self.filter.SetInputData(self.volume_view.actor.imagedata())
        elif slice_type == 1 and self.slice_type != slice_type:
            self.slice_type = slice_type
            self.filter.SetInputData(self.reslice.GetOutput())

    def slice_on_normal(self, origin, normal):
        """
        Slice a volume with a plane oriented by the given normal.
        This allows slicing in all directions.
        :param origin: Origin of the slicing plane
        :param normal: Normal of the slicing plane
        :return: Mesh object with the slice as an image texture
        """
        '''
        mapper = vtk.vtkImageResliceMapper()
        mapper.SetInputData(self.volume_view.actor._data)
        mapper.SliceFacesCameraOff()
        mapper.SliceAtFocalPointOff()
        mapper.JumpToNearestSliceOn()
        mapper.SetImageSampleFactor(2)
        mapper.BorderOn()
        mapper.BackgroundOff()
        mapper.UpdateInformation()
        mapper.GetSlicePlane().SetOrigin(*origin)
        mapper.GetSlicePlane().SetNormal(*normal)
        mapper.GetSlicePlane().Modified()
        mapper.Modified()
        mapper.Update()

        self.actor = vtk.vtkImageSlice()
        self.actor.SetMapper(mapper)
        prop = vtk.vtkImageProperty()
        if True:
            prop.SetInterpolationTypeToLinear()
        else:
            prop.SetInterpolationTypeToNearest()
        self.actor.SetProperty(prop)
        return
        '''
        if self.reslice is None:
            reslice = vtk.vtkImageReslice()
            reslice.SetInputData(self.volume_view.actor._data)
            #reslice.SetInputData(image)
            reslice.SetOutputDimensionality(2)
            reslice.SetAutoCropOutput(False)
            #reslice.SetInterpolationModeToLinear()
            reslice.SetInterpolationModeToNearestNeighbor()
            reslice.SetSlabNumberOfSlices(1)
            reslice.SetOutputSpacing(self.volume_view.get_spacing())
            reslice.ReleaseDataFlagOn()
            self.reslice = reslice
        
        self.set_slice_type(1)
        M, T = utils.get_transformation_matrix(origin, normal)
        self.reslice.SetResliceAxes(M)
        self.reslice.Update()
        self.filter.Update()

        if self.actor is None:
            self.actor = vedo.Mesh(self.filter.GetOutput())
            self.initialize_mapper()
        else:
            self.actor._update(self.filter.GetOutput())

        self.initialize_mapper()

        self.actor.SetOrientation(T.GetOrientation())
        self.actor.SetPosition(origin)
        self.got_slice = True
        return self.actor

    def x_slice(self, i):
        """
        Extract the slice at index `i` of volume along x-axis.
        :param i: I index
        """
        self.set_slice_type(0)
        nx, ny, nz = self.volume_view.actor.GetMapper().GetInput().GetDimensions()
        if i <= 1 or i > nx - 1:
            return False
        self.filter.SetExtent(i, i, 0, ny, 0, nz)
        self.filter.Update()
        if self.actor is not None:
            self.actor._update(self.filter.GetOutput())
        else:
            self.actor = vedo.Mesh(self.filter.GetOutput())
            self.initialize_mapper()
        self.got_slice = True
        return True

    def y_slice(self, j):
        """
        Extract the slice at index `j` of volume along y-axis.
        :param j: J index
        """
        self.set_slice_type(0)
        #nx, ny, nz = self.volume_view.model.dimensions / resolution
        nx, ny, nz = self.volume_view.actor.GetMapper().GetInput().GetDimensions()
        if j <= 1 or j > ny - 1:
            return False
        self.filter.SetExtent(0, nx, j, j, 0, nz)
        self.filter.Update()
        if self.actor is not None:
            self.actor._update(self.filter.GetOutput())
        else:
            self.actor = vedo.Mesh(self.filter.GetOutput())
            self.initialize_mapper()
        self.got_slice = True
        return True

    def z_slice(self, k):
        """
        Extract the slice at index `k` of volume along z-axis.
        :param k: K index
        """
        self.set_slice_type(0)
        nx, ny, nz = self.volume_view.actor.GetMapper().GetInput().GetDimensions()
        if k <= 1 or k > nz - 1:
            return False
        self.filter.SetExtent(0, nx, 0, ny, k, k)
        self.filter.Update()
        if self.actor is not None:
            self.actor._update(self.filter.GetOutput())
        else:
            self.actor = vedo.Mesh(self.filter.GetOutput())
            self.initialize_mapper()
        self.got_slice = True
        return True

    def slice_on_axis(self, value=None, normal=None, axis=None, use_reslice=False):
        """
        Slice on standard X, Y or Z axis
        :param value: Value on the given axis
        :param normal: Axis normal, can be either +1.0 or -1.0 along that axis
        :param axis: Axis integer, 0 for X, 1 for Y, 2 for Z
        :param use_reslice: if True, this enables vtkImageReslice which is useful when
            the normal is not aligned to either X, Y or Z. If you use it on an axis-aligned
            normal, some color inaccuracies will appear if you don't tweak the vtkImageResliceMapper.
            This is why the default is False.
        :return: Result boolean, whether slice occured or not
        """
        resolution = self.volume_view.model.resolution
        volume_dimensions = self.volume_view.model.dimensions
        '''
        if normal[axis] < 0:
            if value > 0:
                # Make value consistent with given normal.
                value *= normal[axis]
            value = volume_dimensions[axis] + value
        '''
        in_volume_slice = int(value) // resolution

        if use_reslice:
            self.slice_on_normal(normal * value, normal)
            return

        if axis == 0:
            result = self.x_slice(in_volume_slice)
        elif axis == 1:
            result = self.y_slice(in_volume_slice)
        elif axis == 2:
            result = self.z_slice(in_volume_slice)
        return result

    def update(self):
        """
        Update slice object according to data in the model
        """
        had_slice = self.got_slice
        result = True
        if isinstance(self.model.axis, int) and 0 <= self.model.axis <= 2:
            result = self.slice_on_axis(self.model.value, self.model.normal, self.model.axis)
        else:
            self.slice_on_normal(self.model.origin, self.model.normal)

        if not result:
            self.plot.remove(self.actor)
            self.got_slice = False
            return

        #self.actor.pos(*(self.volume_view.actor.pos()-self.actor.pos()))
        lut = self.volume_view.model.luts.current
        if lut is not None:
            '''
            This is VTK for you...a mesh can use a vtkLookupTable for RGBA mapping
            BUT volumes require vtkColorTransferFunction (RGB) and vtkPiecewiseFunction (alpha)
            So we have to put a color map, alpha map and a vtkLookupTable 
            built from both maps in a LUTModel.
            Alternatively, we could update the LUT with alpha values but it's a pain.
            
            ctf = self.volume_view.actor.GetProperty().GetRGBTransferFunction()
            lut = vedo.utils.ctf2lut(self.volume_view.actor)
            otf = self.volume_view.actor.GetProperty().GetScalarOpacity

            # using "ctf" would work only for colors, not for transparency!
            self.apply_lut(ctf)
            '''
            self.apply_lut(lut.mapped_lut)
        else:
            if self.alpha_map is None:
                self.actor.cmap(self.color_map)
            else:
                self.actor.cmap(self.color_map, alpha=self.alpha_map)
            
        if self.model.clipping_planes is not None:
            self.actor.mapper().SetClippingPlanes(self.model.clipping_planes)

        if not had_slice:
            self.plot.add(self.actor, render=True)

    def apply_lut(self, lut=None):
        """
        Apply a LUT to the volume
        :param lut: vtkLookupTable
        :param actor: The actor to receive this
        """
        if self.actor is None or lut is None:
            return
        mapper = self.actor._mapper
        mapper.SetLookupTable(lut)


# Add point neurons from connectivity data
from pathlib import Path

import os
import numpy as np
import pickle

from iblviewer.launcher import IBLViewer
from iblviewer import utils

'''
# Invoke your own code with the lines below but you should wrap it in
# properly encapsulated code...
%%bash
git clone https://github.com/int-brain-lab/friendly_neurons.git
%load friendly_neurons/data_analysis.py
'''

# After that, we try loading experimental data, here point neurons.
# Once this cell is run, click on the viewer to see the update

def process_point_neurons(data):
    """
    Pre process point neurons data and find the common min and max values
    :param data: At least 5D array of x, y, z, region, custom data. 
    All the columns after 'region' are be stored as a time series
    """
    if isinstance(data, str):
        pickles = []
        data_path = data
        with (open(os.path.abspath(data_path), 'rb')) as openfile:
            while True:
                try:
                    pickles.append(pickle.load(openfile))
                except EOFError:
                    break
        data = pickles[0]

    # Structure of 'data': x | y | z | region | pre-time allegiance | during-time allegiance | post-time allegiance 
    positions = []
    regions = []
    timings = []
    # Cleaning weird data and extracting what we need. When Alessandro fixes his code, we can get rid of this cleaning.
    for weird_data in data:
        try:
            positions.append([weird_data[0][0], weird_data[1][0], weird_data[2][0]])
        except TypeError:
            continue
            #positions.append(bad_stuff[:3])
        regions.append(weird_data[3])
        timings.append(weird_data[4:])
    positions = np.array(positions).astype(float)

    timings = np.array(timings)
    regions = np.array(regions)
    min_v = np.min(timings)
    max_v = np.max(timings)
    return positions, timings, min_v, max_v
    
def on_viewer_initialized(viewer):
    
    #viewer = MouseBrainViewer()
    #viewer.initialize(resolution=50, mapping='Allen', add_atlas=True, add_dwi=False, dwi_color_map='Greys_r', embed_ui=True)

    # Now add point neurons
    data = ['./exp2_db4df448-e449-4a6f-a0e7-288711e7a75a_both', 
    './exp3_3dd347df-f14e-40d5-9ff2-9c49f84d2157_both', 
    './exp4_3c851386-e92d-4533-8d55-89a46f0e7384_both', 
    './exp5_158d5d35-a2ab-4a76-87b0-51048c5d5283_both']
    data = [str(utils.EXAMPLES_DATA_FOLDER.joinpath(d)) for d in data]

    min_value = None
    max_value = None
    all_positions = []
    all_timings = []
    # A first loop to find out min and max values among timings
    for data_set in data:
        positions, timings, min_v, max_v = process_point_neurons(data_set)
        all_positions.append(positions)
        all_timings.append(timings)
        min_value = min(min_v, min_value) if min_value is not None else min_v
        max_value = max(max_v, max_value) if max_value is not None else max_v
        
    point_actors = []

    # Now visualize the points. Remove existing ones (if any) first and load new ones.
    viewer.plot.remove(point_actors, render=False)
    # We keep newly added objects in memory so that if you rerun this cell and change parameters, 
    # the previous points get replaced by the updated ones. Try changing the radius and run it again.
    point_actors = []
    for d_id in range(len(data)):
        # 16um is a good compromise for visibility from afar. So we make somata roughly 2-3 times larger than they are in reality
        points = viewer.add_points(all_positions[d_id], radius=16, values=all_timings[d_id], screen_space=False, 
                                    noise_amount=100, min_v=min_value, max_v=max_value)
        point_actors.append(points)
    viewer.plot.add(point_actors)


def main():
    iblviewer = IBLViewer()
    # First retrieve command-line arguments (default ones + custom ones above)
    args = iblviewer.parse_args()
    # Now start the viewer and add points when it's initialized
    iblviewer.launch(on_viewer_initialized, None, args)


if __name__ == '__main__':
    main()


try:
    import iblviewer.mouse_brain
    import iblviewer.qt_application
except ModuleNotFoundError:
    pass
import iblviewer.launcher
import iblviewer.collection
import iblviewer.objects
import iblviewer.utils
import iblviewer.volume
import iblviewer.application

if __name__ == '__main__':
    app = iblviewer.launcher.main()


from enum import auto
import sys
from PyQt5 import Qt
from PyQt5 import QtCore
from PyQt5 import QtWidgets
from ipywebrtc.webrtc import VideoStream

# You may need to uncomment these lines on some systems:
#import vtk.qt
#vtk.qt.QVTKRWIBase = "QGLWidget"
from vtk.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import matplotlib as mpl

from iblviewer.application import Viewer
got_ibllib = True
try:
    from iblviewer.mouse_brain import MouseBrainViewer
except ModuleNotFoundError:
    got_ibllib = False
from vedo import Plotter
import numpy as np

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
from threading import *


class MplCanvas(FigureCanvasQTAgg): #or simply FigureCanvas?
    """
    Matplotlib statistics
    """
    def __init__(self, parent=None, width=5, height=4, dpi=100, background_color='white'):
        fig = Figure(figsize=(width, height), dpi=dpi, facecolor=background_color)
        self.axes = fig.add_subplot(111)
        super(MplCanvas, self).__init__(fig)


class ViewerWindow(Qt.QMainWindow):
    """
    This is the main window container that holds the UI, the 3D viewer and statistics
    """
    
    def __init__(self, parent=None):
        """
        Constructor
        """
        Qt.QMainWindow.__init__(self, parent)
        self.frame = Qt.QFrame()

        self.ui_layout = Qt.QVBoxLayout()
        self.main_layout = Qt.QHBoxLayout()

        self.frame.setLayout(self.main_layout)
        self.setCentralWidget(self.frame)

        self.left = 10
        self.top = 10
        self.title = 'International Brain Laboratory Viewer'
        self.width = 1920
        self.height = 1280

        self.viewer = None
        self.viewer_function = None
        self.statistics_function = None
        self.dark_mode = False
        self.kwargs = {}

        self.auto_complete_data_changed = False
        self.regions_search_names = None

        self.setWindowTitle(self.title)
        self.setGeometry(self.left, self.top, self.width, self.height)

        # Overwrite some plot properties to integrate it within Qt
        self.vtkWidget = QVTKRenderWindowInteractor(self.frame)
        self.plot = Plotter(qtWidget=self.vtkWidget)
        self.viewer_initialized = False
        self.neuroscience_context = False
        self.statistics_visible = True

    def initialize(self, viewer=None, callable=None, stats_callable=None, dark_mode=True, **kwargs):
        """
        Initialize the viewer app
        :param viewer: Either iblviewer.application.Viewer or 
            iblviewer.mouse_brain.MouseBrainViewer
        :param callable: A function that you pass to this method. You must write
            that function with a required parameter (the viewer). This allows
            you to perform actions with the viewer before content is shown in QT UI.
            It would also be possible to add a python console in QT in order to run
            Python code live like in Jupyter notebooks but this isn't implemented yet.
        :param stats_callable: A function that will be executed every time the plot is updated
            for instance when a new object or sub selecton is made
        :param dark_mode: Whether the app is in dark mode or not
        """
        self.viewer_function = callable
        self.statistics_function = stats_callable
        self.dark_mode = dark_mode
        if isinstance(kwargs, dict):
            self.kwargs = kwargs
        if got_ibllib:
            if isinstance(viewer, MouseBrainViewer) or isinstance(viewer, Viewer):
                self.viewer = viewer
            if self.viewer is None:
                self.viewer = MouseBrainViewer()
            self.neuroscience_context = isinstance(self.viewer, MouseBrainViewer)
        else:
            self.neuroscience_context = False

        self.initialize_ui()

        # It's important to start the viewer in another thread so that the QT UI
        # doesn't freeze when interacting with it.
        # In case you need to go further with this, look at QThread with a good summary here:
        # https://realpython.com/python-pyqt-qthread/#using-qthread-to-prevent-freezing-guis
        #thread = Thread(target=self._initialize_viewer)
        #thread.start()
        self._initialize_viewer()

    def _initialize_viewer(self):
        """
        Initialize the viewer
        """
        #if viewer.plot is not None:
            #viewer.plot.close()
        if 'embed_ui' in self.kwargs:
            self.kwargs['embed_ui'] = False
        #if 'offscreen' in self.kwargs:
            #self.kwargs['offscreen'] = True
        if self.viewer is None:
            if got_ibllib:
                self.viewer = MouseBrainViewer()
            else:
                self.viewer = Viewer()

        self.viewer.initialize(plot=self.plot, dark_mode=self.dark_mode, **self.kwargs)
        if self.viewer_function is not None:
            try:
                # Allows users to add data to be visualized
                self.viewer_function(self.viewer)
            except Exception as e:
                print(e)
        
        # Assign functions (to mimick event callbacks)
        # Have a look at Qt signals if you want to do something more advanced
        # Basic example : https://stackoverflow.com/questions/28793440/pyqt5-focusin-out-events
        self.viewer.selection_changed = self.onSelectionChanged
        self.viewer.sub_selection_changed = self.onSelectionChanged
        self.viewer.objects_changed = self.onObjectsChanged

        self.viewer_initialized = True
        self.show_viewer()

    def initialize_ui(self):
        """
        Initialize the QT UI
        """
        self.background_color = 'white'
        mpl.rcParams['axes.spines.left'] = True
        mpl.rcParams['axes.spines.right'] = False
        mpl.rcParams['axes.spines.top'] = False
        mpl.rcParams['axes.spines.bottom'] = True
        self.set_dark_mode(self.dark_mode)
        
        self.main_layout.addLayout(self.ui_layout, 1)

        self.menu_tabs = Qt.QTabWidget()
        self.camera_menu = Qt.QWidget()
        self.tools_menu = Qt.QWidget()
        self.object_menu = Qt.QWidget()
        #self.data_menu = Qt.QWidget()
        self.export_menu = Qt.QWidget()
        self.menu_tabs.addTab(self.camera_menu, 'Camera')
        self.menu_tabs.addTab(self.tools_menu, 'Tools')
        self.menu_tabs.addTab(self.object_menu, 'Object')
        #self.menu_tabs.addTab(self.data_menu, 'Data')
        self.menu_tabs.addTab(self.export_menu, 'Export')

        #camera_group = QtWidgets.QGroupBox('Camera presets')
        self.camera_menu.layout = Qt.QVBoxLayout()
        hbox = Qt.QHBoxLayout()
        self.add_button('Left', self.onLeftCameraPreset, hbox)
        self.add_button('Right', self.onRightCameraPreset, hbox)
        self.camera_menu.layout.addLayout(hbox)
        hbox = Qt.QHBoxLayout()
        self.add_button('Dorsal', self.onDorsalCameraPreset, hbox)
        self.add_button('Ventral', self.onVentralCameraPreset, hbox)
        self.camera_menu.layout.addLayout(hbox)
        hbox = Qt.QHBoxLayout()
        self.add_button('Anterior', self.onAnteriorCameraPreset, hbox)
        self.add_button('Posterior', self.onPosteriorCameraPreset, hbox)
        self.camera_menu.layout.addLayout(hbox)
        hbox = Qt.QHBoxLayout()
        self.add_checkbox('Autofocus', self.onToggleAutofocus, hbox, set_checked=True)
        self.add_checkbox('Orthographic', self.onToggleOrthoCamera, hbox)
        hbox.addStretch(1)
        self.camera_menu.layout.addLayout(hbox)
        hbox = Qt.QHBoxLayout()
        self.add_button('View selected', self.onViewSelection, hbox)
        self.camera_menu.layout.addLayout(hbox)
        self.camera_menu.layout.addStretch(1)
        self.camera_menu.setLayout(self.camera_menu.layout)
        #self.ui_layout.addWidget(camera_group)

        scene_group = QtWidgets.QGroupBox('Scene')
        vbox = Qt.QVBoxLayout()
        self.selection_combo = self.add_combo('Select an object', [], self.onChangeSelection, vbox)
        scene_group.setLayout(vbox)

        view_group = QtWidgets.QGroupBox('View options')
        vbox = Qt.QVBoxLayout()
        hbox = Qt.QHBoxLayout()
        self.add_checkbox('Outline', self.onToggleOutline, hbox, set_checked=True)
        self.add_checkbox('Marker', self.onToggleMarker, hbox, set_checked=True)
        self.add_checkbox('Color bar', self.onToggleColorBar, hbox, set_checked=True)
        vbox.addLayout(hbox)
        hbox = Qt.QHBoxLayout()
        self.add_checkbox('Information text', self.onToggleInfoText, hbox, set_checked=True)
        self.add_checkbox('Dark background', self.onToggleDarkBackground, hbox, set_checked=True)
        vbox.addLayout(hbox)
        hbox = Qt.QHBoxLayout()
        self.add_checkbox('Fast volumes', self.onToggleVolumesLOD, hbox)
        vbox.addLayout(hbox)
        view_group.setLayout(vbox)

        self.ui_layout.addWidget(scene_group)
        self.ui_layout.addWidget(view_group)

        if self.statistics_function is not None:
            stats_group = QtWidgets.QGroupBox('Statistics')
            hbox = Qt.QHBoxLayout()
            self.add_checkbox('Statistics panel', self.onToggleStatistics, hbox, auto_render=False)
            stats_group.setLayout(hbox)
            self.ui_layout.addWidget(stats_group)

        #object_group = QtWidgets.QGroupBox('Object settings')
        self.object_menu.layout = Qt.QVBoxLayout()
        self.opacity_slider = self.add_slider('Opacity', 1.0, 0.0, 1.0, 0.05, self.onOpacityChange, self.object_menu.layout)
        self.slices_opacity_slider = self.add_slider('Slices opacity', 1.0, 0.0, 1.0, 0.05, 
                                                    self.onSliceOpacityChange, self.object_menu.layout)
        self.hollow_checkbox = self.add_checkbox('Hollow regions', self.onToggleHollow, self.object_menu.layout)
        self.time_series_slider = self.add_slider('Time series', 1, 0, 10, 1, self.onTimeSeriesChange, self.object_menu.layout)
        self.add_button('Remove object', self.onRemoveObject, self.object_menu.layout)
        self.object_menu.layout.addStretch(1)
        self.object_menu.setLayout(self.object_menu.layout)
        #self.ui_layout.addWidget(object_group)
        #self.object_menu.layout.addWidget(object_group)

        self.tools_menu.layout = Qt.QVBoxLayout()
        self.new_probe_button = self.add_button('Add new probe', self.onNewProbe, self.tools_menu.layout)
        self.edit_probe_button = self.add_button('Edit probe', self.onEditProbe, self.tools_menu.layout)
        self.slicer_button = self.add_button('Cutter/Slicer mode', self.onSlicerToggle, self.tools_menu.layout, toggle=True)

        if self.neuroscience_context:
            self.search_input = self.add_input('Search an atlas region', None, self.tools_menu.layout, autocomplete=True)
            #self.isosurface_checkbox = self.add_checkbox('Show regions surface', None, vbox, set_checked=True)
            self.search_button = self.add_button('Get region', self.onSearch, self.tools_menu.layout)

        self.clipping_combo = self.add_combo('Select a clipping object', [], None, self.tools_menu.layout, auto_render=False)
        self.clip_to_bounds = self.add_button('Clip to object bounds', self.onClipToBounds, self.tools_menu.layout)
        self.tools_menu.layout.addStretch(1)
        self.tools_menu.setLayout(self.tools_menu.layout)

        #group = QtWidgets.QGroupBox('Data settings')
        #self.data_menu.layout = Qt.QVBoxLayout()
        #self.selection_label = QtWidgets.QLabel()
        #self.data_menu.layout.addWidget(self.selection_label)
        #self.data_menu.layout.addStretch(1)
        #self.data_menu.setLayout(self.data_menu.layout)
        #self.ui_layout.addWidget(data_group)

        #video_group = QtWidgets.QGroupBox('Video export')
        self.export_menu.layout = Qt.QVBoxLayout()
        video_group = QtWidgets.QGroupBox('Export video presets')
        vbox = Qt.QVBoxLayout()
        self.duration_slider = self.add_slider('Duration', 8, 0, 60, 1, None, vbox)
        self.start_angle_slider = self.add_slider('Start angle', 0, 0, 360, 1, None, vbox)
        self.end_angle_slider = self.add_slider('End angle', 360, 0, 360, 1, None, vbox)
        self.end_angle_slider.setTracking(True)
        self.end_angle_slider.setTickPosition(QtWidgets.QSlider.TicksAbove)
        self.export_button = self.add_button('Export video...', self.onExportVideo, vbox)
        video_group.setLayout(vbox)
        #self.video_menu.layout.addStretch(1)
        self.export_menu.layout.addWidget(video_group)

        image_group = QtWidgets.QGroupBox('Export image presets')
        vbox = Qt.QVBoxLayout()
        self.magnification_scale = self.add_slider('Magnification scale', 2, 1, 10, 1, None, vbox)
        self.export_button = self.add_button('Export image...', self.onExportImage, vbox)
        image_group.setLayout(vbox)
        self.export_menu.layout.addStretch(1)
        self.export_menu.layout.addWidget(image_group)

        self.export_menu.setLayout(self.export_menu.layout)
        #self.ui_layout.addWidget(video_group)
        
        # Once we're done with preparing the whole menu with tabs, add that to the UI
        self.ui_layout.addWidget(self.menu_tabs)
        self.ui_layout.addStretch(1)

        splitter = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        splitter.addWidget(self.vtkWidget)
        
        self.statistics = None
        self.windows_qt_mpl_issue = sys.platform == "win32"
        if not self.windows_qt_mpl_issue:
            self.statistics = MplCanvas(self, 5, 4, 100, self.background_color)
            self.statistics.setStyleSheet("background-color:transparent;")
            self.update_statistics()
            # Create toolbar, passing statistics as first parament, parent (self, the MainWindow) as second.
            #toolbar = NavigationToolbar(self.statistics, self)
            #layout = Qt.QVBoxLayout()
            #layout.addWidget(toolbar)
            #layout.addWidget(self.statistics)
            #self.main_vbox.addLayout(layout)

            self.statistics_widget = Qt.QWidget()
            self.statistics_layout = Qt.QHBoxLayout(self.statistics_widget)
            self.statistics_layout.addWidget(self.statistics)
            # Don't show stats initially
            self.onToggleStatistics()

            splitter.addWidget(self.statistics_widget)
        
        self.main_layout.addWidget(splitter, 4)
        self.main_splitter = splitter
        #splitter.setStretchFactor(3, 1)
        #splitter.setSizes([125, 150])
        
        self.vtkWidget.update()
        self.show()

    def onViewerInitialized(self):
        if self.neuroscience_context:
            regions_data = self.viewer.get_region_names()
            self.search_input.completer_model.setStringList(regions_data)

    def show_viewer(self):
        """
        Show the viewer (when it's initialized)
        """
        self.viewer.show()
        self.vtkWidget.update()
        self.onObjectsChanged()
        self.onViewerInitialized()
        self.update_ui()

    def set_light_mode(self):
        """
        Set light mode to viewer and stats
        """
        self.set_dark_mode(False)

    def set_dark_mode(self, on=True):
        """
        Set dark mode to viewer and stats
        :param on: Whether dark mode is on
        """
        if on:
            self.background_color = '#2d2d2d'
            plt.style.use('dark_background')
            mpl.rcParams['axes.facecolor'] = self.background_color
        else:
            self.background_color = '#dddddd'
            plt.style.use('default')
            mpl.rcParams['axes.facecolor'] = self.background_color

        if self.viewer is not None:
            self.viewer.set_dark_mode(on)
    
    def update_ui(self):
        """
        Update the QT UI
        """
        got_selection = self.viewer.model.got_selection()
        is_volume = self.viewer.is_volume()
        is_probe = self.viewer.is_probe()

        self.new_probe_button.setEnabled(got_selection and is_volume)
        self.edit_probe_button.setEnabled(got_selection and is_probe)
        #self.isosurface_checkbox.setEnabled(got_selection and is_volume)
        self.hollow_checkbox.setEnabled(got_selection and is_volume)

        self.clipping_combo.setEnabled(got_selection)
        self.clip_to_bounds.setEnabled(got_selection)
        self.slicer_button.setEnabled(got_selection)

        self.slicer_button.setChecked(self.viewer.box_widget is not None)
        
        if got_selection:
            '''
            text, data = self.viewer.get_selection_info(line_length=32)
            if text is not None:
                self.selection_label.setText(text+'\n')
            else:
                self.selection_label.setText('')
            '''
            index = self.selection_combo.findText(self.viewer.model.selection.name)
            if index != -1:
                self.selection_combo.setCurrentIndex(index)
        else:
            self.selection_combo.setCurrentIndex(0)
            #self.selection_label.setText('')
            
        opacity = self.viewer.get_selection_opacity()
        self.opacity_slider.setEnabled(opacity is not None)
        self.opacity_slider.label.setEnabled(opacity is not None)
        '''
        if opacity is None:
            self.opacity_slider.label.hide()
        else:
            self.opacity_slider.show()
            self.opacity_slider.label.show()
        '''
        if opacity is not None:
            self.opacity_slider.setValue(int(opacity / self.opacity_slider.step))

        slices_opacity = self.viewer.get_selection_slices_opacity()
        self.slices_opacity_slider.setEnabled(slices_opacity is not None)
        self.slices_opacity_slider.label.setEnabled(slices_opacity is not None)
        if slices_opacity is not None:
            self.slices_opacity_slider.setValue(int(slices_opacity / self.slices_opacity_slider.step))
        
        current_value, max_value = self.viewer.get_time_series_info()
        self.time_series_slider.setEnabled(current_value is not None)
        if isinstance(current_value, int):
            self.time_series_slider.setValue(current_value)
        if isinstance(max_value, int):
            self.time_series_slider.setRange(0, max_value)

    def update_statistics(self):
        """
        Update statistics
        """
        if not self.statistics_visible:
            return
        if self.statistics_function is not None:
            try:
                plot = self.statistics_function(self.statistics, self.viewer)
                if plot is not None:
                    self.statistics = plot
            except Exception as e:
                e_type = type(e).__name__
                msg = 'Your custom statistics function failed with error type ' + e_type
                if 'TypeError' in e_type:
                    msg += '\nMake sure your functions accepts two arguments (the statistics plot and the 3d viewer).'
                print(msg)
                print(e)
        else:
            self.statistics.axes.clear()
            self.statistics.axes.plot(np.arange(20), np.random.random(20)/2)
            #self.statistics.setStyleSheet("background-color:#eeeeee;")
            self.statistics.draw()

    def onObjectsChanged(self):
        """
        Event triggered when the dictionary of 3D objects has been updated
        """
        names = self.viewer.get_object_names()
        current_id = None
        for n_id in range(len(names)):
            if names[n_id] == self.selection_combo.currentText():
                current_id = n_id
        names = ['None'] + names
        self.selection_combo.clear()
        self.selection_combo.addItems(names)
        self.clipping_combo.clear()
        self.clipping_combo.addItems(names)
        if current_id is not None:
            self.selection_combo.setCurrentIndex(current_id)

    def onChangeSelection(self, value):
        """
        Event triggered by QT to change the viewer's selection
        """
        index = self.selection_combo.findText(value)
        if index != -1:
            self.selection_combo.setCurrentIndex(index)
        self.viewer.select(value, allow_none=True)

    def onSelectionChanged(self):
        """
        Event triggered by the viewer when a new valid selection is made
        """
        self.update_ui()
        self.update_statistics()

    def onSearch(self):
        search_term = self.search_input.text()
        if search_term == '':
            return
        result = self.viewer.find_region(search_term)
        if len(result) < 1:
            return
        # TODO: improve this and give user the choice (left or right hemisphere)
        region_id = result[0]
        #if self.isosurface_checkbox.isChecked():
        self.viewer.isosurface(region_id, split_meshes=False)

    def onOpacityChange(self, value):
        self.viewer.update_opacity(value=value)

    def onSliceOpacityChange(self, value):
        self.viewer.update_slices_opacity(value=value)

    def onTimeSeriesChange(self, value):
        self.viewer.set_time_series(value)

    @Qt.pyqtSlot()
    def onToggleStatistics(self):
        if self.windows_qt_mpl_issue:
            return
        if self.statistics_visible:
            self.statistics_visible = False
            self.statistics_widget.hide()
        else:
            self.statistics_visible = True

            # Using show on the maptlotlib widget will segfault with an error
            # in mpl backend wrt to QT. So it's better to start with a new plot
            self.statistics_layout.removeWidget(self.statistics)
            self.statistics = MplCanvas(self, 5, 4, 100, self.background_color)
            self.statistics.setStyleSheet("background-color:transparent;")
            self.update_statistics()
            self.statistics_layout.addWidget(self.statistics)
            self.statistics_widget.show()

    @Qt.pyqtSlot()
    def onToggleMarker(self):
        self.viewer.toggle_marker()

    @Qt.pyqtSlot()
    def onToggleOutline(self):
        self.viewer.toggle_outline()

    @Qt.pyqtSlot()
    def onToggleColorBar(self):
        self.viewer.toggle_color_bar()

    @Qt.pyqtSlot()
    def onToggleInfoText(self):
        self.viewer.toggle_info_text()

    @Qt.pyqtSlot()
    def onToggleDarkBackground(self):
        self.viewer.toggle_dark_mode()

    @Qt.pyqtSlot()
    def onToggleVolumesLOD(self):
        self.viewer.toggle_volumetric_lod()

    @Qt.pyqtSlot()
    def onToggleMarkerType(self):
        self.viewer.toggle_marker_type()

    @Qt.pyqtSlot()
    def onToggleHollow(self):
        self.viewer.toggle_hollow_mode()

    @Qt.pyqtSlot()
    def onRemoveObject(self):
        self.viewer.remove_object()
        self.onObjectsChanged()
            
    @Qt.pyqtSlot()
    def onExportImage(self):
        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.DontUseNativeDialog
        file_path, _ = QtWidgets.QFileDialog.getSaveFileName(self, 'QFileDialog.getSaveFileName()', 
                                                'iblviewer.jpg', 'All Files (*);;Images (*.png *.jpg);', 
                                                options=options)
        if file_path:
            self.viewer.render(file_path)
            Qt.QMessageBox.about(self, 'Image rendering complete', f'File was saved under {file_path}')
            
    @Qt.pyqtSlot()
    def onExportVideo(self):
        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.DontUseNativeDialog
        file_path, _ = QtWidgets.QFileDialog.getSaveFileName(self, 'QFileDialog.getSaveFileName()', 
                                                'iblviewer.mp4', 'All Files (*);;Videos (*.mp4);', 
                                                options=options)
        if file_path:
            start_angle = self.start_angle_slider.value()
            end_angle = self.end_angle_slider.value()
            duration = self.duration_slider.value()
            self.viewer.export_turntable_video(file_path, start_angle, end_angle, duration)
            Qt.QMessageBox.about(self, 'Video rendering complete', f'File was saved under {file_path}')
    
    @Qt.pyqtSlot()
    def onClipToBounds(self):
        bounds_obj = self.viewer.objects.get(self.clipping_combo.currentText())
        self.viewer.clip_to_bounds(bounds_obj)

    @Qt.pyqtSlot()
    def onIsosurface(self):
        self.viewer.isosurface(split_meshes=True)

    @Qt.pyqtSlot()
    def onSlicerToggle(self):
        self.viewer.toggle_box_widget()

    @Qt.pyqtSlot()
    def onNewProbe(self):
        self.viewer.add_probe()

    @Qt.pyqtSlot()
    def onEditProbe(self):
        self.viewer.edit_probe()

    @Qt.pyqtSlot()
    def onToggleOrthoCamera(self):
        self.viewer.toggle_orthographic_view()

    @Qt.pyqtSlot()
    def onViewSelection(self):
        self.viewer.view_selected()

    @Qt.pyqtSlot()
    def onToggleAutofocus(self):
        self.viewer.toggle_autofocus()

    @Qt.pyqtSlot()
    def onLeftCameraPreset(self):
        self.viewer.set_left_view()

    @Qt.pyqtSlot()
    def onRightCameraPreset(self):
        self.viewer.set_right_view()

    @Qt.pyqtSlot()
    def onDorsalCameraPreset(self):
        self.viewer.set_dorsal_view()

    @Qt.pyqtSlot()
    def onVentralCameraPreset(self):
        self.viewer.set_ventral_view()

    @Qt.pyqtSlot()
    def onAnteriorCameraPreset(self):
        self.viewer.set_anterior_view()

    @Qt.pyqtSlot()
    def onPosteriorCameraPreset(self):
        self.viewer.set_posterior_view()

    @Qt.pyqtSlot()
    def onClose(self):
        #Disable the interactor before closing to prevent it
        #from trying to act on already deleted items
        self.vtkWidget.close()

    # Below are utility functions 

    def add_input(self, label_text, change_function=None, layout=None, autocomplete=False):
        """
        Add an input line with a label
        """
        label = QtWidgets.QLabel(self)
        label.setText(label_text)
        input = QtWidgets.QLineEdit(self)
        input.label = label
        if change_function is not None:
            input.editingFinished.connect(change_function)
        if autocomplete:
            completer = QtWidgets.QCompleter()
            model = Qt.QStringListModel()
            completer.setModel(model)
            completer.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
            completer.setCompletionMode(QtWidgets.QCompleter.InlineCompletion)
            input.setCompleter(completer)
            # Store properties in order to update the model with a list later on
            input.completer_model = model
            input.completer = completer
            # If autocomplete is a list of strings, we use it as the completion list
            if isinstance(autocomplete, list):
                try:
                    input.completer_model.setStringList(autocomplete)
                except Exception:
                    pass
        if layout is not None:
            layout.addWidget(label)
            layout.addWidget(input)
        return input

    def add_button(self, label, click_function, layout=None, tooltip=None, 
                    auto_render=True, toggle=False, set_checked=False):
        """
        Add a new button to a layout
        """
        button = QtWidgets.QPushButton(label, self)
        if isinstance(tooltip, str):
            button.setToolTip(tooltip)
        def click_handler(value):
            if click_function is not None:
                click_function()
            if auto_render:
                self.viewer.render()
        button.setCheckable(toggle)
        if toggle and set_checked:
            button.setChecked(set_checked)
        button.clicked.connect(click_handler)
        if layout is not None:
            layout.addWidget(button)
        return button

    def add_checkbox(self, label, click_function, layout=None, tooltip=None, 
                    auto_render=True, set_checked=False):
        """
        Add a new checkbox to a layout
        """
        checkbox = QtWidgets.QCheckBox(label, self)
        checkbox.move(20, 0)
        if isinstance(tooltip, str):
            checkbox.setToolTip(tooltip)
        def change_handler(value):
            if click_function is not None:
                click_function()
            if auto_render:
                self.viewer.render()
        checkbox.setChecked(set_checked)
        checkbox.stateChanged.connect(change_handler)
        if layout is not None:
            layout.addWidget(checkbox)
        return checkbox

    def add_combo(self, text, values=None, change_function=None, layout=None, auto_render=True):
        """
        Add a new combobox with a label to a layout
        """
        label = QtWidgets.QLabel(self)
        label.setText(text)

        combo = QtWidgets.QComboBox(self)
        combo.label = label
        combo.last_value = None
        if values is not None:
            combo.addItems(values)
        
        def update_combo(value):
            #label.setText(f'{value}')
            if change_function is not None:
                change_function(value)
            if auto_render:
                self.viewer.render()
            combo.last_value = value

        if change_function is not None:
            combo.activated[str].connect(update_combo)
        #combo.currentIndexChanged['QString'].connect(update_combo)
        #combo.valueChanged.connect(update_combo)
        
        if layout is not None:
            layout.addWidget(label)
            layout.addWidget(combo)
        return combo

    def add_slider(self, text, value, min_value=0, max_value=10, step=1, change_function=None, 
                    layout=None, ticks=None, label_precision=2, auto_render=True):
        """
        Add a new slider with a label to a layout
        """
        label = QtWidgets.QLabel(self)
        label.setText(text)
        #label.setPixmap(QPixmap('mute.png'))

        slider = QtWidgets.QSlider(QtCore.Qt.Horizontal, self)
        #slider.setGeometry(50,50, 200, 50)
        #slider.move(0, 30)
        # Sliders in Qt can only accept integers (!)
        # so we have to work around that
        if max_value <= 1.0 and step < 1.0:
            min_value /= step
            max_value /= step
            value /= step
        # Custom dynamic properties
        slider.label = label
        slider.step = step
        slider.last_value = None

        slider.setMinimum(int(min_value))
        slider.setMaximum(int(max_value))
        slider.setValue(int(value))
        slider.setMinimumWidth(200)

        real_value = value
        if slider.step < 1.0:
            real_value = value * step
        if isinstance(real_value, float):
            label.setText(f'{text}: {real_value:.{label_precision}}')
        else:
            label.setText(f'{text}: {real_value}')

        if ticks is None:
            ticks = {'interval':2, 'position':'below'}
        #slider.setFocusPolicy(Qt.StrongFocus)
        slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        if ticks.get('interval') is not None:
            slider.setTickInterval(ticks.get('interval'))
        if step is not None:
            slider.setSingleStep(max(1, step))
        
        def update_slider_label(value):
            real_value = value
            if slider.step < 1.0:
                real_value = value * step
            if isinstance(real_value, float):
                label.setText(f'{text}: {real_value:.{label_precision}}')
            else:
                label.setText(f'{text}: {real_value}')
            if change_function is not None:
                change_function(real_value)
            if auto_render:
                self.viewer.render()
            slider.last_value = real_value
        
        slider.valueChanged.connect(update_slider_label)
        
        if layout is not None:
            layout.addWidget(label)
            layout.addWidget(slider)
        return slider


class ViewerApp(Qt.QApplication):
    """
    This is the main entry point to start a Qt application.
    """
    def __init__(self):
        super().__init__(sys.argv)
        self.window = ViewerWindow()
        self.aboutToQuit.connect(self.window.onClose)

    def initialize(self, viewer=None, callable=None, stats_callable=None, dark_mode=True, **kwargs):
        if dark_mode:
            # Other dark-style Qt stylesheets exist but few get it right
            from darktheme.widget_template import DarkPalette
            self.setStyle('Fusion')
            self.setPalette(DarkPalette())
            # Handling disabled states with the custom stylesheet below
            self.setStyleSheet("QToolTip { color: #ffffff; background-color: grey; border: 1px solid white; }"
            "QCheckBox:disabled {color:#999999;}"
            "QSlider::sub-page:disabled {background:#999999;}"
            "QRadioButton:disabled {color:#999999;}"
            "QWidget:disabled {color:#999999;}")
        Qt.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
        self.window.initialize(viewer, callable, stats_callable, dark_mode, **kwargs)
        Qt.QApplication.restoreOverrideCursor()
        self.exec_()

def main():
    app = ViewerApp(dark_mode=False)

if __name__ == "__main__":
    main()


from iblviewer.mouse_brain import MouseBrainViewer


def main():
    # This example starts the viewer and renders a view
    viewer = MouseBrainViewer()
    # Font size is made larger for 4K rendering
    viewer.initialize(resolution=50, mapping='Allen', embed_font_size=30, offscreen=True)
    # Put objects on the scene
    viewer.show()

    # Add more code to add data and control the viewer here

    # Render a 4K image.
    viewer.set_info_visibility(False)
    #viewer.render('./test.jpg', 3840, 2160)
    viewer.close()


if __name__ == '__main__':
    main()


from setuptools import setup, find_packages
#import iblviewer.launcher

setup(
    name='iblviewer',
    version='2.4.7',
    description='An interactive GPU-accelerated 3D viewer based on VTK',
    url='https://github.com/int-brain-lab/iblviewer',
    author='Nicolas Antille',
    author_email='nicolas.antille@gmail.com',
    license='MIT',
    install_requires=['numpy',
                      'matplotlib',
                      'requests',
                      'pandas',
                      'pynrrd',
                      'trimesh',
                      'k3d',
                      'vtk>=9.0',
                      'ipywebrtc',
                      'ibllib',
                      'iblutil',
                      'vedo>=2022.0.1',
                      'ipyvtklink',
                      'PyQt5',
                      'pyqt-darktheme'
                      ],
    packages=find_packages(include=['iblviewer','iblviewer.*', 'iblviewer_examples', 'iblviewer_examples.*']),
    package_data={'iblviewer_assets':['iblviewer_assets/*']},
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "iblviewer = iblviewer.launcher:main",
            "iblviewer-points-demo = iblviewer_examples.ibl_point_neurons:main",
            "iblviewer-probes-demo = iblviewer_examples.ibl_insertion_probes:main",
            "iblviewer-coverage-demo = iblviewer_examples.ibl_brain_coverage:main",
            "iblviewer-volume-mapping-demo = iblviewer_examples.ibl_volume_mapping:main",
            "iblviewer-brain-wide-map = iblviewer_examples.ibl_brain_wide_map:main",
            "iblviewer-human-brain-demo = iblviewer_examples.human_brain:main",
            "iblviewer-headless-render-demo = iblviewer_examples.headless_render:main"
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3.8',
    ]
)
